/****************************************************************************************
*
*    File Name:  model_rldram3.v
*      Version:  1.20
*        Model:  BUS Functional
*
*  Description:  Micron RLDRAM3
*
*   Limitation:  - 
*
*         Note:  - Set simulator resolution to "ps" accuracy
*                - Set DEBUG = 0 to disable $display messages
*
*   Disclaimer   This software code and all associated documentation, comments or other
*  of Warranty:  information (collectively "Software") is provided "AS IS" without
*                warranty of any kind. MICRON TECHNOLOGY, INC. ("MTI") EXPRESSLY
*                DISCLAIMS ALL WARRANTIES EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
*                TO, NONINFRINGEMENT OF THIRD PARTY RIGHTS, AND ANY IMPLIED WARRANTIES
*                OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE. MTI DOES NOT
*                WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE
*                OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE.
*                FURTHERMORE, MTI DOES NOT MAKE ANY REPRESENTATIONS REGARDING THE USE OR
*                THE RESULTS OF THE USE OF THE SOFTWARE IN TERMS OF ITS CORRECTNESS,
*                ACCURACY, RELIABILITY, OR OTHERWISE. THE ENTIRE RISK ARISING OUT OF USE
*                OR PERFORMANCE OF THE SOFTWARE REMAINS WITH YOU. IN NO EVENT SHALL MTI,
*                ITS AFFILIATED COMPANIES OR THEIR SUPPLIERS BE LIABLE FOR ANY DIRECT,
*                INDIRECT, CONSEQUENTIAL, INCIDENTAL, OR SPECIAL DAMAGES (INCLUDING,
*                WITHOUT LIMITATION, DAMAGES FOR LOSS OF PROFITS, BUSINESS INTERRUPTION,
*                OR LOSS OF INFORMATION) ARISING OUT OF YOUR USE OF OR INABILITY TO USE
*                THE SOFTWARE, EVEN IF MTI HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
*                DAMAGES. Because some jurisdictions prohibit the exclusion or
*                limitation of liability for consequential or incidental damages, the
*                above limitation may not apply to you.
*
*                Copyright 2009 Micron Technology, Inc. All rights reserved.
*
* Rev   Author   Date        Changes
* ---------------------------------------------------------------------------------------
* 0.00  MYY      10/07/09    Initial Release
* 0.10  MYY      04/09/10    DM 2 bit; WL=RL+1
* 0.20  MYY      04/09/10    QK 4 bit
* 0.30  MYY      05/25/10    Updated Timing Error Messages
* 1.00  MYY      02/07/11    Updated to design sheet v1.00
* 1.10  MYY      07/27/11    Updated tCK calculation for clock jitter
* 1.11  MYY      03/07/12    Support parameterized part width
* 1.12  MYY      03/05/13    Added initialization check
* 1.13  MYY      03/18/13    Align read DQ/QVLD with QK 
* 1.14  MYY      04/02/13    Increase DQ/QVLD margin for clock jitter 
* 1.15  MYY      04/25/13    Align RTR timing with Read timing 
* 1.16  MYY      05/23/13    Write data capture on DK0/1
* 1.17  MYY      08/07/13    1Gb support
* 1.18  MYY      04/17/14    Use DQ_Model_In for write data collection
* 1.19  MYY      06/03/14    SIO support
* 1.20  MYY      08/03/15    Support full range tCKDK
****************************************************************************************/


// DO NOT CHANGE THE TIMESCALE
// MAKE SURE YOUR SIMULATOR USES "PS" RESOLUTION
`timescale 1ps / 1ps

module model_rldram3 (
                    CK,
                    CS_Model_,
                    
                    WE_,
                    REF_,
                    ADDR,
                    BA,
                    
                    DQ_Model,

                    DK,
                    DK_,
                    DM,

                    QVLD_Model,
                    QK_Model,
                    QK_Model_,

                    ZQ,
                    RESET,
                    VREF,
                    MF,
                    
                    // JTAG pins
                    TMS,
                    TDI,
                    TCK,
                    TDO_Model

                    );

`ifdef TB
   parameter MAX_DQ_BITS = 36;
   parameter MAX_QK_BITS = 4;
`else // extern model use
 `ifdef x18
   parameter MAX_DQ_BITS = 18;
   parameter MAX_QK_BITS = 2;
 `else
   parameter MAX_DQ_BITS = 36;
   parameter MAX_QK_BITS = 4;
 `endif
`endif

   import rldram3_package::*;

   // driver array
   reg [iMAXPINS:0] pindrv;
   reg [iMAXPINS:0] tapdrv;
   reg              tapon;

   input [1:0]     CK;
   input [1:0]     CS_Model_;
   input           WE_;
   input           REF_;
   input [MAX_ADDR_BITS - 1 : 0] ADDR;
   input [MAX_BA_BITS - 1 : 0]   BA;

   inout [MAX_DQ_BITS - 1 : 0]   DQ_Model;
   assign                        DQ_Model = tapon ? tapdrv[iDQin35:iDQin0] : pindrv[iDQin35:iDQin0];
   reg [35 : 0]                 DQ_Model_In;
   assign                       DQ_Model_In[MAX_DQ_BITS - 1 : 0] = DQ_Model;
    
    
   input [MAX_DK_BITS - 1 : 0]   DK;
   input [MAX_DK_BITS - 1 : 0]   DK_;
   input [MAX_DM_BITS - 1 : 0]   DM;
   
   output [MAX_QVLD_BITS - 1 : 0] QVLD_Model;
   wire [MAX_QVLD_BITS - 1 : 0]   QVLD_Model_Int;
   output [MAX_QK_BITS - 1 : 0]   QK_Model;
   output [MAX_QK_BITS - 1 : 0]   QK_Model_;
   assign                        (supply0, supply1) QVLD_Model_Int = tapon ? tapdrv[iQVLD1:iQVLD0] : pindrv[iQVLD1:iQVLD0];
   assign                        (supply0, supply1) #(0,200) QVLD_Model = QVLD_Model_Int;
   assign                        (supply0, supply1) QK_Model = tapon ? tapdrv[iQK3:iQK0] : pindrv[iQK3:iQK0];
   assign                        (supply0, supply1) QK_Model_ = tapon ? tapdrv[iQK3F:iQK0F] : pindrv[iQK3F:iQK0F];
   
   output                        TDO_Model;
   assign                        (supply0, supply1) TDO_Model = pindrv[iTDO];
   
   input                         ZQ;
   input                         RESET;
   input                         VREF;
   input                         MF;
   
   input                         TMS;
   input                         TDI;
   input                         TCK;

   // ODT val for each pin
   reg [2:0]                     DQ_ODT_Model [MAX_DQ_BITS - 1 : 0];
   reg [2:0]                     DM_ODT_Model [MAX_DM_BITS - 1 : 0];
   reg [2:0]                     DK_ODT_Model [MAX_DK_BITS - 1 : 0];
   // DK_ same state as DK
   // QVLD/QK has ODT in M/S mode
   reg [2:0]                     QVLD_ODT_Model [MAX_QVLD_BITS - 1 : 0];
   reg [2:0]                     QK_ODT_Model [MAX_QK_BITS - 1 : 0];
   // QK_ same state as QK

   // ODI val for each pin
   reg [1:0]                     DQ_ODI_Model [MAX_DQ_BITS - 1 : 0];
   reg [1:0]                     QK_ODI_Model [MAX_QK_BITS - 1 : 0];
   // QK_ same state as QK
   reg [1:0]                     QVLD_ODI_Model [MAX_QVLD_BITS - 1 : 0];

   integer                        Burst_Length, Burst_Type;
   integer                        Data_Width, ODT_val, ODI_val;
   integer                        Write_Latency;
   integer                        Read_Latency;

   // only link to other values for backward compatibility
   assign                         Burst_Length = modeconfig.BL;
   assign                         Burst_Type = 0;
   assign                         Read_Latency = modeconfig.DL;
   assign                         Write_Latency = modeconfig.DL+1;
   assign                         Data_Width = dutconfig.bymode;                                         
   assign                         ODT_val = modeconfig.ODT;
   assign                         ODI_val = modeconfig.ODI;
   
   // save last sampled cmd addr
   UTYPE_cmdpacket cmdpkt,curcmdpkt,lastcmdpkt;
   reg [MAX_ADDR_BITS - 1 : 0]    lastADDR;
   reg [MAX_BA_BITS - 1 : 0]      lastBA;

   time                           arraywritetime;
   reg [40*8-1:0]                 msg;
  
   // event to process write
   reg                            e_process_write;
   reg [MAX_ADDR_BITS - 1 : 0]    writeaddr;
   reg [MAX_BA_BITS - 1 : 0]      writebank;

   // event to process read
   reg                            e_process_read;
   reg [MAX_ADDR_BITS - 1 : 0]    readaddr;
   reg [MAX_BA_BITS - 1 : 0]      readbank;
   
   // timing
   UTYPE_sharedtiming timing;

   // dut config
   UTYPE_dutconfig dutconfig;
   
   // dut state
   UTYPE_dutstate dutstate;
   
   // dut mode config
   UTYPE_dutmodeconfig modeconfig;

   time                           prevposedgeCK, tm_reset0;
   int                            resetdone, cy_reset0, cy_reset5, cy_reset8;
   int                            resetqk;
   bit                            GenExpData, waitclkr;
   int tCK, tOffset;

   parameter                      PERTCKAVG=10;
   
   // clock jitter
   real                           tck_avg;
   time                           tck_sample [PERTCKAVG-1:0];
   integer                        tck_i;
   integer                        ck_cntr;
  
   // state of the chip
   statetable #(.MAX_DQ_BITS(MAX_DQ_BITS),
                .MAX_QK_BITS(MAX_QK_BITS)) 
     IState(.dutconfig(dutconfig),
            .modeconfig(modeconfig),
            .dutstate(dutstate)
            );
   
   // setup dutconfigs
`ifdef TB
   // when run with test bench
   assign                         dutconfig = `TB.dutconfig;
`else
   
`ifdef rldram3_1Gb
 `include "timing_tasks_1Gb.v"
`else // 576Mb
 `include "timing_tasks.v"
`endif
   
   // run with model only
   initial 
     begin
`ifdef DDP
        dutconfig.archtype = RLDRAM3DDP;
`else
        dutconfig.archtype = RLDRAM3;
`endif

        if (MAX_DQ_BITS == 18) begin
           dutconfig.config_bymode = BYX18;
           dutconfig.bymode = BYX18;
        end
        else begin
           dutconfig.config_bymode = BYX36;
           dutconfig.bymode = BYX36;
        end
        dutconfig_populate(dutconfig);

`ifdef sgtap    
        SetPatGenTiming(TS_TAP);
`elsif sg83f
        SetPatGenTiming(TS_83F);
`elsif sg83e
        SetPatGenTiming(TS_83E);
`elsif sg93f
        SetPatGenTiming(TS_93F);
`elsif sg93e
        SetPatGenTiming(TS_93E);
`elsif sg93
        SetPatGenTiming(TS_93);
`elsif sg107f
        SetPatGenTiming(TS_107F);
`elsif sg107e
        SetPatGenTiming(TS_107E);
`elsif sg107
        SetPatGenTiming(TS_107);
`elsif sg125e
        SetPatGenTiming(TS_125E);
`else 
        SetPatGenTiming(TS_125);
`endif 
        updatetiming(timing);
     end
   
   function void dutconfig_populate (inout UTYPE_dutconfig dutconfig);
         dutconfig.numbank = 16;
         dutconfig.bankaddrmask = 4'b1111;
`ifdef SIO
	 dutconfig.sio=1;
`else
	 dutconfig.sio=0;
`endif	 

`ifdef rldram3_1Gb         
         case (dutconfig.bymode)
           BYX36: begin
              dutconfig.addrmask = 22'hf_ffff;
              dutconfig.numaddr = 20;
              dutconfig.dqmask = 36'hf_ffff_ffff;
              dutconfig.qkmask = 4'hf;
              dutconfig.numdq = 36;
              dutconfig.numqk = 4;
              dutconfig.numdm = 2;
              dutconfig.numqvld = 2;

              // DDP 
              if (dutconfig.archtype == RLDRAM3DDP) begin
                 dutconfig.addrmask = 22'h1f_ffff;
                 dutconfig.numaddr = 21;
                 dutconfig.ddpranked=0;
              end
           end
           BYX18: begin
              dutconfig.addrmask = 22'h1f_ffff;
              dutconfig.numaddr = 21;
              dutconfig.dqmask = 36'h0_0003_ffff;
              dutconfig.qkmask = 4'h3;
              dutconfig.numdq = 18;
              dutconfig.numqk = 2;
              dutconfig.numdm = 2;
              dutconfig.numqvld = 1;

              // DDP
              if (dutconfig.archtype == RLDRAM3DDP) begin
                 dutconfig.addrmask = 22'h3f_ffff;
                 dutconfig.numaddr = 22;
                 dutconfig.ddpranked=1;
                 if (dutconfig.sio) begin
                     dutconfig.numqk = 4;
                     dutconfig.numqvld = 2;
                 end
              end
           end
         endcase
         dutconfig.JTAG_IDCODE = 32'b0001_0001000110100111_00000101100_1;
`else // 576Mb
         case (dutconfig.bymode)
           BYX36: begin
              dutconfig.addrmask = 21'h7_ffff;
              dutconfig.numaddr = 19;
              dutconfig.dqmask = 36'hf_ffff_ffff;
              dutconfig.qkmask = 4'hf;
              dutconfig.numdq = 36;
              dutconfig.numqk = 4;
              dutconfig.numdm = 2;
              dutconfig.numqvld = 2;

              // DDP 
              if (dutconfig.archtype == RLDRAM3DDP) begin
                 dutconfig.addrmask = 21'hf_ffff;
                 dutconfig.numaddr = 20;
                 dutconfig.ddpranked=0;
              end
           end
           BYX18: begin
              dutconfig.addrmask = 21'hf_ffff;
              dutconfig.numaddr = 20;
              dutconfig.dqmask = 36'h0_0003_ffff;
              dutconfig.qkmask = 4'h3;
              dutconfig.numdq = 18;
              dutconfig.numqk = 2;
              dutconfig.numdm = 2;
              dutconfig.numqvld = 1;

              // DDP
              if (dutconfig.archtype == RLDRAM3DDP) begin
                 dutconfig.addrmask = 21'h1f_ffff;
                 dutconfig.numaddr = 21;
                 dutconfig.ddpranked=1;
              end
           end
         endcase
         dutconfig.JTAG_IDCODE = 32'b0001_0001000110100111_00000101100_1;
`endif
   endfunction

`endif

   // func clocks
   always @(CK[1]) begin: cmd_loop
      reg [MAX_DQ_BITS*MAX_BURST_LEN-1:0] data;
      reg [MAX_BA_BITS - 1 : 0]           bank;
      reg [MAX_ADDR_BITS - 1 : 0]         addr;
            
      // posedge
      if (CK[1] === 1'b1 && waitclkr === 1'b1) begin: cmdloop

         #1;
         
         // update tCK
         if (resetdone==2) begin
            tck_i = $time - prevposedgeCK;
            tck_avg = tck_avg - tck_sample[ck_cntr%PERTCKAVG]/$itor(PERTCKAVG);
            tck_avg = tck_avg + tck_i/$itor(PERTCKAVG);
            tck_sample[ck_cntr%PERTCKAVG] = tck_i;
            tCK = $rtoi(tck_avg);

            if (timing.tOffset !== tCK) begin
               timing.tOffset = tCK;
            end
            
            timing.tCK=tCK;
         end  
         
         ck_cntr <= ck_cntr + 1;
         prevposedgeCK = $time;
         
         cmdpkt = IState.cmddecode(RESET, CS_Model_, WE_, REF_, lastcmdpkt, modeconfig);
         cmdpkt.tCK=tCK;
         cmdpkt.addr = ADDR;
         if(dutconfig.ddpranked) cmdpkt.addr[dutconfig.numaddr-1] = cmdpkt.cs;
         cmdpkt.bank = BA;
         cmdpkt.addryflag = 1'b0; // no y part in cmdpkt
         curcmdpkt = cmdpkt;

     // initialization check
     if (resetdone==0) begin
        cy_reset0++;
     end
         if (resetdone==1) begin // need 1 stable clock
`ifndef SKIPINITCHK
        if ($time-tm_reset0 < 200_000_000ps) $display("MODEL_ERROR @%t %m INIT3 violation, <200us RESET low ",$time);
        if (cy_reset0 < 100) $display("MODEL_ERROR @%t %m INIT4 violation, <100 cycles RESET low ",$time);
`endif
        cy_reset5=0;
            resetdone=2;
         end
`ifndef SKIPINITCHK
         if (resetdone==2) begin 
        if (cmdpkt.cmd==cmdNOP && cmdpkt.bank[3:0]===4'h0 && cmdpkt.addr[18:0]===19'h0)
          cy_reset5++;
        else begin
          if (cy_reset5 < 10_000) $display("MODEL_ERROR @%t %m INIT5 violation, 10_000 NOPs with BA/ADDR low ",$time);
          resetdone=3;
        end
         end
         if (resetdone==3) begin 
        if ( (cmdpkt.cmd==cmdLMR || cmdpkt.cmd==cmdLMRx) && cmdpkt.bank[3:0]===4'h0)
          resetdone=4;
     end    
         if (resetdone==4) begin 
            if (modeconfig.AM) begin
           if ( cmdpkt.cmd==cmdLMRy && lastcmdpkt.cmd==cmdLMRx && lastcmdpkt.bank[3:0]==4'b1) begin
              if (lastcmdpkt.addr[5]!==1'b1) $display("MODEL_ERROR @%t %m INIT7 violation, DLLRESET required",$time);
              if (cmdpkt.addr[8]!==1'b1) $display("MODEL_ERROR @%t %m INIT7 violation, ZQINIT required",$time);
              resetdone=5;
           end
        end
        else begin
           if (cmdpkt.cmd==cmdLMR && BA[3:0]===4'h1) begin
              if (cmdpkt.addr[5]!==1'b1) $display("MODEL_ERROR @%t %m INIT7 violation, DLLRESET required",$time);
              if (cmdpkt.addr[6]!==1'b1) $display("MODEL_ERROR @%t %m INIT7 violation, ZQINIT required",$time);
              resetdone=5;
           end
        end
     end    
         if (resetdone==5 && cmdpkt.cmd==cmdNOP) begin 
        cy_reset8=0;
        resetdone=6;
     end
         if (resetdone==6) begin 
        if (cmdpkt.cmd==cmdNOP)
          cy_reset8++;
        else begin
           if (dutconfig.archtype==RLDRAM3DDP && dutconfig.bymode== BYX36 && cy_reset8<1024)
             $display("MODEL_ERROR @%t %m INIT8 violation, 1024 NOPs after ZQCL ",$time);
           else if (cy_reset8<512)
             $display("MODEL_ERROR @%t %m INIT8 violation, 512 NOPs after ZQCL ",$time);
           resetdone=7;
        end
         end
         if (resetdone==7) begin 
        if ( (cmdpkt.cmd==cmdLMR || cmdpkt.cmd==cmdLMRx) && BA[3:0]===4'h2)
          resetdone=99; // reset done
     end
     if (resetdone!=99 && 
         (curcmdpkt.cmd == cmdRD || curcmdpkt.cmd == cmdRDx || curcmdpkt.cmd == cmdWR || curcmdpkt.cmd == cmdWRx ||
          curcmdpkt.cmd == cmdRT || curcmdpkt.cmd == cmdRTx || curcmdpkt.cmd == cmdREF1 || curcmdpkt.cmd == cmdREF1x ||
          curcmdpkt.cmd == cmdREFM || curcmdpkt.cmd == cmdREFMx)) begin
        $display("MODEL_ERROR @%t %m invalid command before INIT is done",$time);
        resetdone=99; // reset done
     end
`endif
         
         if (modeconfig.AM) begin
            if (curcmdpkt.cmd == cmdRDy && lastcmdpkt.cmd != cmdRDx) $display("MODEL_ERROR @%t %m Read command Addr Mux mode",$time);
            if (curcmdpkt.cmd == cmdRTy && lastcmdpkt.cmd != cmdRTx) $display("MODEL_ERROR @%t %m ReadTrain command Addr Mux mode",$time);
            if (curcmdpkt.cmd == cmdWRy && lastcmdpkt.cmd != cmdWRx) $display("MODEL_ERROR @%t %m ReadTrain command Addr Mux mode",$time);
            if (curcmdpkt.cmd == cmdLMRy && lastcmdpkt.cmd != cmdLMRx) $display("MODEL_ERROR @%t %m ReadTrain command Addr Mux mode",$time);
            if (curcmdpkt.cmd == cmdREF1y && lastcmdpkt.cmd != cmdREF1x) $display("MODEL_ERROR @%t %m ReadTrain command Addr Mux mode",$time);
            if (curcmdpkt.cmd == cmdREFMy && lastcmdpkt.cmd != cmdREFMx) $display("MODEL_ERROR @%t %m ReadTrain command Addr Mux mode",$time);
         end 

         // check comand timing
         if ( IState.minnops(cmdpkt,msg)>0) 
           $display("@%t %m MODEL_ERROR Timing Error %s",$time, msg);
         
         // update to table
         IState.updatetable(0, curcmdpkt, -1);

         // now modeconfig are updated
         // set ODT
         if (curcmdpkt.cmd == cmdLMRy ||
             curcmdpkt.cmd == cmdLMR) begin
            // ODT 
            if (dutconfig.archtype == RLDRAM3DDP) begin
               // ODT not modeled in DDP
            end
            else begin
               for (int i=0; i<dutconfig.numdq; i++)
                 DQ_ODT_Model[i] <= #(timing.tOffset) modeconfig.ODT;
               for (int i=0; i<MAX_DK_BITS; i++)
                 DK_ODT_Model[i] <= #(timing.tOffset) modeconfig.ODT;
               for (int i=0; i<dutconfig.numdm; i++)
                 DM_ODT_Model[i] <= #(timing.tOffset) modeconfig.ODT;
               // ODS
               for (int i=0; i<dutconfig.numdq; i++)
                 DQ_ODI_Model[i] <= #(timing.tOffset) 'z;
               for (int i=0; i<dutconfig.numqk; i++)
                 QK_ODI_Model[i] <= #(timing.tOffset) modeconfig.ODI;
               for (int i=0; i<dutconfig.numqvld; i++)
                 QVLD_ODI_Model[i] <= #(timing.tOffset) modeconfig.ODI;
            end

            // start driving qk
            resetqk=0;
         end
         
         // read train
         if (curcmdpkt.cmd == cmdRT ||
             curcmdpkt.cmd == cmdRTy) begin
            
        fork begin
           #250;
           IState.readtraindata(data);
               // read train is always BL4
               //LoadBurstExpDQ(modeconfig.DL-1, modeconfig.BL, data);
               //LoadBurstExpQVLD(modeconfig.DL-1, modeconfig.BL);
               LoadBurstExpDQ(modeconfig.DL-1, BSTL4, data, 0);
               LoadBurstExpQVLD(modeconfig.DL-1, BSTL4, '1);
        end
        join_none
         end
         
         // reads are scheduled after command are decoded, it is not blocked even in case of timing violations
         if (curcmdpkt.cmd == cmdRD ||
             curcmdpkt.cmd == cmdRDy) begin
            if (modeconfig.AM && curcmdpkt.cmd == cmdRDy) begin
               IState.MuxXY2A(lastcmdpkt.addr, curcmdpkt.addr, addr);
               bank = lastcmdpkt.bank;
            end
            else begin
               addr = curcmdpkt.addr;
               bank = curcmdpkt.bank;
            end
            
            e_process_read <= #(tCK*(modeconfig.DL-2) +250) 1'b1;
            e_process_read <= #(tCK*(modeconfig.DL-2)+2 +250) 1'b0;
            readbank <= #(tCK*(modeconfig.DL-2) -2 +250) bank;
            readaddr <= #(tCK*(modeconfig.DL-2) -2 +250) addr;
         end

         // writes are scheduled after command are decoded, it is not blocked even in case of timing violations
         if (curcmdpkt.cmd == cmdWR ||
             curcmdpkt.cmd == cmdWRy) begin
            if (modeconfig.AM && curcmdpkt.cmd == cmdWRy)  begin
               IState.MuxXY2A(lastcmdpkt.addr, curcmdpkt.addr, addr);
               bank = lastcmdpkt.bank;
            end
            else begin
               addr = curcmdpkt.addr;
               bank = curcmdpkt.bank;
            end
            
            e_process_write <= #(tCK*(modeconfig.DL+1+modeconfig.BL/2) - (tCK*20/100)) 1'b1;
            e_process_write <= #(tCK*(modeconfig.DL+1+modeconfig.BL/2) -(tCK*20/100) + 2) 1'b0;
            writebank <= #(tCK*(modeconfig.DL+1+modeconfig.BL/2)  -2 -(tCK*20/100)) bank;
            writeaddr <= #(tCK*(modeconfig.DL+1+modeconfig.BL/2)  -2 -(tCK*20/100)) addr;
         end
         
         // load QK
         if (resetqk==0) begin
            int tCKQK;
            if (modeconfig.DLL===1'b1)
              tCKQK=timing.tCKQK_DLL_DIS;
            else 
              tCKQK=timing.tCKQK;
            setpinCK(iQK0, tCKQK, 2'b01);
            setpinCK(iQK0F, tCKQK, 2'b10);
            setpinCK(iQK1, tCKQK, 2'b01);
            setpinCK(iQK1F, tCKQK, 2'b10);
            if (4 == dutconfig.numqk) begin
               setpinCK(iQK2, tCKQK, 2'b01);
               setpinCK(iQK2F, tCKQK, 2'b10);
               setpinCK(iQK3, tCKQK, 2'b01);
               setpinCK(iQK3F, tCKQK, 2'b10);
            end
         end
         
         // prepare next cycle
         lastcmdpkt = cmdpkt;
         lastADDR = ADDR;
         lastBA = BA;
         
         if (resetdone) begin
            // wait for falling edge
            waitclkr=1'b0;
         end  
      end
      
      // negedge
      if (CK[1] === 1'b0 && waitclkr === 1'b0) begin
         #1;
         
         // wait for rising edge
         waitclkr=1'b1;
      end
   end

   reg [36*(MAX_BURST_LEN+2)-1:0] wrdata_r, wrdata;
   reg [MAX_DM_BITS*(MAX_BURST_LEN+2)-1:0] dmdata;
   
   // collect write data
   always @(DK[0]) begin
      if (DK[0]==1'b1) begin
         if (MAX_DQ_BITS == 18) begin
                wrdata_r[18+8:18+0] =  DQ_Model_In[8:0];
         end else begin
                wrdata_r[36+8:36+0] =  DQ_Model_In[8:0];
                wrdata_r[36+26:36+18] =  DQ_Model_In[26:18];
         end
             dmdata[2] = DM[0];
      end
      else if (DK[0]==1'b0) begin
         if (MAX_DQ_BITS == 18) begin
                wrdata[18+8:18+0] = wrdata_r[18+8:18+0];
                wrdata[8:0] =  DQ_Model_In[8:0];
         end else begin
                wrdata[36+8:36+0] = wrdata_r[36+8:36+0];
                wrdata[36+26:36+18] = wrdata_r[36+26:36+18];
                wrdata[8:0] =  DQ_Model_In[8:0];
                wrdata[26:18] =  DQ_Model_In[26:18];
         end
             dmdata[0] = DM[0];

         if (MAX_DQ_BITS == 18) begin
            wrdata[36*4+26:36*4+18]=wrdata[36*3+26:36*3+18];
            wrdata[36*4+ 8:36*4+ 0]=wrdata[36*3+ 8:36*3+ 0];
            wrdata[36*3+26:36*3+18]=wrdata[36*2+26:36*2+18];
            wrdata[36*3+ 8:36*3+ 0]=wrdata[36*2+ 8:36*2+ 0];
            wrdata[36*2+26:36*2+18]=wrdata[36*1+26:36*1+18];
            wrdata[36*2+ 8:36*2+ 0]=wrdata[36*1+ 8:36*1+ 0];
            wrdata[36*1+26:36*1+18]=wrdata[36*0+26:36*0+18];
            wrdata[36*1+ 8:36*1+ 0]=wrdata[36*0+ 8:36*0+ 0];
         end else begin
            wrdata[72*4+62:72*4+54]=wrdata[72*3+62:72*3+54];
            wrdata[72*4+44:72*4+36]=wrdata[72*3+44:72*3+36];
            wrdata[72*4+26:72*4+18]=wrdata[72*3+26:72*3+18];
            wrdata[72*4+ 8:72*4+ 0]=wrdata[72*3+ 8:72*3+ 0];
            wrdata[72*3+62:72*3+54]=wrdata[72*2+62:72*2+54];
            wrdata[72*3+44:72*3+36]=wrdata[72*2+44:72*2+36];
            wrdata[72*3+26:72*3+18]=wrdata[72*2+26:72*2+18];
            wrdata[72*3+ 8:72*3+ 0]=wrdata[72*2+ 8:72*2+ 0];
            wrdata[72*2+62:72*2+54]=wrdata[72*1+62:72*1+54];
            wrdata[72*2+44:72*2+36]=wrdata[72*1+44:72*1+36];
            wrdata[72*2+26:72*2+18]=wrdata[72*1+26:72*1+18];
            wrdata[72*2+ 8:72*2+ 0]=wrdata[72*1+ 8:72*1+ 0];
            wrdata[72*1+62:72*1+54]=wrdata[72*0+62:72*0+54];
            wrdata[72*1+44:72*1+36]=wrdata[72*0+44:72*0+36];
            wrdata[72*1+26:72*1+18]=wrdata[72*0+26:72*0+18];
            wrdata[72*1+ 8:72*1+ 0]=wrdata[72*0+ 8:72*0+ 0];
         end

         dmdata[4*4+2] = dmdata[4*3+2];
         dmdata[4*4+0] = dmdata[4*3+0];
         dmdata[4*3+2] = dmdata[4*2+2];
         dmdata[4*3+0] = dmdata[4*2+0];
         dmdata[4*2+2] = dmdata[4*1+2];
         dmdata[4*2+0] = dmdata[4*1+0];
         dmdata[4*1+2] = dmdata[4*0+2];
         dmdata[4*1+0] = dmdata[4*0+0];
      end
   end

   always @(DK[1]) begin
      if (DK[1]==1'b1) begin
         if (MAX_DQ_BITS == 18) begin
                wrdata_r[18+17:18+9] =  DQ_Model_In[17:9];
         end else begin
                wrdata_r[36+17:36+9] =  DQ_Model_In[17:9];
                wrdata_r[36+35:36+27] =  DQ_Model_In[35:27];
         end
             dmdata[3] = DM[1];
      end
      else if (DK[1]==1'b0) begin
         if (MAX_DQ_BITS == 18) begin
                wrdata[18+17:18+9] = wrdata_r[18+17:18+9];
                wrdata[17:9] =  DQ_Model_In[17:9];
         end else begin
                wrdata[36+17:36+9] = wrdata_r[36+17:36+9];
                wrdata[36+35:36+27] = wrdata_r[36+35:36+27];
                wrdata[17:9] =  DQ_Model_In[17:9];
                wrdata[35:27] =  DQ_Model_In[35:27];
         end
             dmdata[1] = DM[1];

         if (MAX_DQ_BITS == 18) begin
            wrdata[36*4+35:36*4+27]=wrdata[36*3+35:36*3+27];
            wrdata[36*4+17:36*4+ 9]=wrdata[36*3+17:36*3+ 9];
            wrdata[36*3+35:36*3+27]=wrdata[36*2+35:36*2+27];
            wrdata[36*3+17:36*3+ 9]=wrdata[36*2+17:36*2+ 9];
            wrdata[36*2+35:36*2+27]=wrdata[36*1+35:36*1+27];
            wrdata[36*2+17:36*2+ 9]=wrdata[36*1+17:36*1+ 9];
            wrdata[36*1+35:36*1+27]=wrdata[36*0+35:36*0+27];
            wrdata[36*1+17:36*1+ 9]=wrdata[36*0+17:36*0+ 9];
         end else begin
            wrdata[72*4+71:72*4+63]=wrdata[72*3+71:72*3+63];
            wrdata[72*4+53:72*4+45]=wrdata[72*3+53:72*3+45];
            wrdata[72*4+35:72*4+27]=wrdata[72*3+35:72*3+27];
            wrdata[72*4+17:72*4+ 9]=wrdata[72*3+17:72*3+ 9];
            wrdata[72*3+71:72*3+63]=wrdata[72*2+71:72*2+63];
            wrdata[72*3+53:72*3+45]=wrdata[72*2+53:72*2+45];
            wrdata[72*3+35:72*3+27]=wrdata[72*2+35:72*2+27];
            wrdata[72*3+17:72*3+ 9]=wrdata[72*2+17:72*2+ 9];
            wrdata[72*2+71:72*2+63]=wrdata[72*1+71:72*1+63];
            wrdata[72*2+53:72*2+45]=wrdata[72*1+53:72*1+45];
            wrdata[72*2+35:72*2+27]=wrdata[72*1+35:72*1+27];
            wrdata[72*2+17:72*2+ 9]=wrdata[72*1+17:72*1+ 9];
            wrdata[72*1+71:72*1+63]=wrdata[72*0+71:72*0+63];
            wrdata[72*1+53:72*1+45]=wrdata[72*0+53:72*0+45];
            wrdata[72*1+35:72*1+27]=wrdata[72*0+35:72*0+27];
            wrdata[72*1+17:72*1+ 9]=wrdata[72*0+17:72*0+ 9];
         end

         dmdata[4*4+3] = dmdata[4*3+3];
         dmdata[4*4+1] = dmdata[4*3+1];
         dmdata[4*3+3] = dmdata[4*2+3];
         dmdata[4*3+1] = dmdata[4*2+1];
         dmdata[4*2+3] = dmdata[4*1+3];
         dmdata[4*2+1] = dmdata[4*1+1];
         dmdata[4*1+3] = dmdata[4*0+3];
         dmdata[4*1+1] = dmdata[4*0+1];
      end
   end
   
   // read to array
   always @(posedge e_process_read) begin: _n_e_process_read
      reg [MAX_DQ_BITS*MAX_BURST_LEN-1:0] data;
      reg [MAX_QVLD_BITS - 1 : 0]        qvldmask;
      int sioshift;

      qvldmask='1;
      sioshift=0;
      if (dutconfig.archtype==RLDRAM3DDP && dutconfig.ddpranked && dutconfig.sio && readaddr[dutconfig.numaddr-1]) begin      
          // DDP RANKED SIO, read data shift from 0:17 to 35:18 for die1
          sioshift=18;
          // toggle qvld[1]
          qvldmask=2'b10;
      end
      else if (dutconfig.archtype==RLDRAM3DDP && dutconfig.ddpranked && dutconfig.sio && readaddr[dutconfig.numaddr-1]==1'b0) begin      
          // toggle qvld[0]
          qvldmask=2'b01;
      end
      
      IState.readarray(readbank,readaddr,data,arraywritetime);
      if (modeconfig.DLL===1'b1) begin
         // DLL_DIS data came out 1 cycle earlier
         LoadBurstExpDQ(0, modeconfig.BL, data, sioshift);
         LoadBurstExpQVLD(0, modeconfig.BL, qvldmask);
      end
      else begin
         LoadBurstExpDQ(1, modeconfig.BL, data, sioshift);
         LoadBurstExpQVLD(1, modeconfig.BL, qvldmask);
      end        
   end
   
   // write to array
   always @(posedge e_process_write) begin: _n_e_process_write
      reg [MAX_DQ_BITS*MAX_BURST_LEN-1:0] data;
      reg [MAX_DM_BITS*MAX_BURST_LEN-1:0] dm;
      
      for(int i=modeconfig.BL-1;i>=0; i--)
        for(int j=0; j<MAX_DQ_BITS; j++)
          if (dutconfig.archtype==RLDRAM3 && dutconfig.sio) begin 
              // SDP SIO, write data shift from 35:18 to 17:0 
              if(dutconfig.dqmask[j])
                data[i*MAX_DQ_BITS+j] = wrdata[(modeconfig.BL-i-1+2)*MAX_DQ_BITS+j+18];
          end
          else if (dutconfig.archtype==RLDRAM3DDP && dutconfig.ddpranked && dutconfig.sio && writeaddr[dutconfig.numaddr-1]==1'b0) begin 
              // DDP RANKED SIO, write data shift from 35:18 to 17:0 for die0
              if(dutconfig.dqmask[j])
                data[i*MAX_DQ_BITS+j] = wrdata[(modeconfig.BL-i-1+2)*MAX_DQ_BITS+j+18];
          end
          else begin
              if(dutconfig.dqmask[j])
                data[i*MAX_DQ_BITS+j] = wrdata[(modeconfig.BL-i-1+2)*MAX_DQ_BITS+j];
          end
      
      for(int i=modeconfig.BL-1;i>=0; i--)
        for(int j=0; j<MAX_DM_BITS; j++)
          dm[i*MAX_DM_BITS+j] = dmdata[(modeconfig.BL-i-1+2)*MAX_DM_BITS+j];
      
      IState.writearray(writebank,writeaddr,data,dm);
     end
   
   // reset done
   always @(posedge RESET) begin
      resetdone=1;
   end
   
   // reset begin
   always @(negedge RESET) begin
      resetdone=0;
      resetqk=1;
      modelreset;
    end
   
   // initilize
   initial begin
      modelreset;
      waitclkr=1'b1;
      tapon=1'b0;
      e_process_write=1'b0;
      e_process_read=1'b0;
   end
   
   // tasks
   task modelreset;
      begin
         // init varibles
         prevposedgeCK = 0;
         resetdone=0;
         tm_reset0=$time;
         cy_reset0=0;
         tCK=0;
             ck_cntr=0;
             GenExpData=1;
         
         PatGenPwrReset;
      end
   endtask

   task PatGenPwrReset;
      begin
         UTYPE_cmdpacket cmdpkt;

         // init state table
         cmdpkt.cmd=cmdINIT;
         cmdpkt.tCK=5000;
         IState.updatetable(GenExpData, cmdpkt, -1);
         
         // init pins
         pindrv = '1;
         pindrv[iDQ35:iDQ0] = 'z;
         pindrv[iDQin35:iDQin0] = 'z;
         pindrv[iQVLD1:iQVLD0] = 'z;
         pindrv[iQK3:iQK0] = 'z;
         pindrv[iQK3F:iQK0F] = 'z;

         // reset again after 20ns
         pindrv[iDQ35:iDQ0] <= #20ns 'z;
         pindrv[iDQin35:iDQin0] <= #20ns 'z;
         pindrv[iQVLD1:iQVLD0] <= #20ns 'z;
         pindrv[iQK3:iQK0] <= #20ns 'z;
         pindrv[iQK3F:iQK0F] <= #20ns 'z;

         // drive QVLD to 0 
         for (int i=0; i<dutconfig.numqvld; i++) begin
            pindrv[iQVLD0+i] = 1'b0;
            pindrv[iQVLD0+i] <= #(20ns+1ps) 1'b0;
         end
         
         // init ODT
         for (int i=0; i<MAX_DQ_BITS; i++)
           DQ_ODT_Model[i] = 0;
         for (int i=0; i<MAX_DM_BITS; i++)
           DM_ODT_Model[i] = 0;
         for (int i=0; i<MAX_DK_BITS; i++)
           DK_ODT_Model[i] = 0;
         for (int i=0; i<MAX_QVLD_BITS; i++)
           QVLD_ODT_Model[i] = 0;
         for (int i=0; i<dutconfig.numqvld; i++)
           QVLD_ODT_Model[i] = 'z;
         for (int i=0; i<MAX_QK_BITS; i++)
           QK_ODT_Model[i] = 0;
         for (int i=0; i<dutconfig.numqk; i++)
           QK_ODT_Model[i] = 'z;

         // init ODI
         for (int i=0; i<MAX_DQ_BITS; i++)
           DQ_ODI_Model[i] = 'z;
         for (int i=0; i<MAX_QVLD_BITS; i++)
           QVLD_ODI_Model[i] = 'z;
         for (int i=0; i<MAX_QK_BITS; i++)
           QK_ODI_Model[i] = 'z;
      end
   endtask

   task setpinCK (int i, int delay, reg [MAX_PINV-1:0] val);
      pindrv[i] <= #(timing.tOffset + delay) val[0];
      pindrv[i] <= #(timing.tOffset + timing.tCK*timing.ClockDutyCycle/100 + delay) val[1];
   endtask
   
   task setpinCA (int i, int ts, reg val);
      pindrv[i] <= #(timing.tOffset - ts) val;
   endtask 

   task setpinDQr (int i, int lat, int delay, int ts, reg val );
      pindrv[i] <= #(timing.tOffset + lat*timing.tCK + delay - ts) val;
   endtask
   
   task setpinDQf (int i, int lat, int delay, int ts, reg val );
      pindrv[i] <= #(timing.tOffset + lat*timing.tCK + timing.tCK*timing.ClockDutyCycle/100 + delay - ts) val;
   endtask
   
   task LoadBurstExpQVLD;
      input                                    int dl,bl;
      input [MAX_QVLD_BITS - 1 : 0]            mask;
      begin: LoadBurstExpQVLD_func
         int tCKQK;
         if (modeconfig.DLL===1'b1) 
           tCKQK=timing.tCKQK_DLL_DIS;
         else 
           tCKQK=timing.tCKQK;
         for (int i=0; i<dutconfig.numqvld; i++)
           if (mask[i]) begin
               setpinDQf(iQVLD0+i, dl-1, tCKQK, 250, 1'b1);
               setpinDQf(iQVLD0+i, dl-1+bl/2, tCKQK, 400, 1'b0);
           end
      end
   endtask

   task LoadBurstExpDQ;
      input                                    int dl,bl;
      input [MAX_DQ_BITS*MAX_BURST_LEN-1:0]    data ;
      input                                    int sioshift;
      begin: LoadBurstExpDQ_func
         int tCKQK;
         if (modeconfig.DLL===1'b1)
           tCKQK=timing.tCKQK_DLL_DIS;
         else 
           tCKQK=timing.tCKQK;

         // set burst
         for (int l=0; l<bl; l+=2) 
           for (int i=0; i<dutconfig.numdq; i++) begin
              if (dutconfig.dqmask[i]) begin
                 // set DQ
                 setpinDQr(iDQin0+i+sioshift, dl+l/2, tCKQK, 250, data[l*MAX_DQ_BITS+i]);
                 setpinDQf(iDQin0+i+sioshift, dl+l/2, tCKQK, 250, data[(l+1)*MAX_DQ_BITS+i]);
              end
              // odt
              if (dutconfig.archtype == RLDRAM3DDP) begin
                 // ODT not modelled in DDP
              end
              else begin
                 if (dutconfig.dqmask[i]) begin
                    // set ODT to z, ODT N/A as it is driving
                    DQ_ODT_Model[i] <= #(timing.tOffset + (dl+l/2)*timing.tCK + tCKQK - 250 ) 'z;
                    // set ODI 
                    DQ_ODI_Model[i] <= #(timing.tOffset + (dl+l/2)*timing.tCK + tCKQK - 250) modeconfig.ODI;
                 end
              end                
           end 
         

         // set burst end
         for (int i=0; i<dutconfig.numdq; i++) begin
            if (dutconfig.dqmask[i]) begin
               // set DQ to z
               setpinDQr(iDQin0+i+sioshift, dl+bl/2, tCKQK, 350, 1'bz);
            end
            // ODT
            if (dutconfig.archtype == RLDRAM3DDP) begin
            end
            else begin
               if (dutconfig.dqmask[i]) begin
                  // set ODT to mode reg value
                  DQ_ODT_Model[i] <= #(timing.tOffset + (dl+bl/2)*timing.tCK + tCKQK - 350) modeconfig.ODT;
                  // set ODI to z
                  DQ_ODI_Model[i] <= #(timing.tOffset + (dl+bl/2)*timing.tCK + tCKQK - 350) 'z;
               end
            end          
         end 
         
      end
   endtask

   // TAP functions
   reg tdo, tdo_valid;
   bit [BSR_LEN-1:0] TAPBSRtemp;
   bit [BSR_LEN_EXT-1:0] TAPBSREXTtemp;
   always @(TCK) begin
      #3ps;

      if (TCK===1'b1) begin
         
         // sample pins for TAP
         TAPBSRtemp[IState.TAPpin2BSR(iCK)] = CK[1];
         TAPBSRtemp[IState.TAPpin2BSR(iCKF)] = CK[0];
         TAPBSRtemp[IState.TAPpin2BSR(iCSF0)] = CS_Model_[0];
         TAPBSRtemp[IState.TAPpin2BSR(iWEF)] = WE_;
         TAPBSRtemp[IState.TAPpin2BSR(iREFF)] = REF_;
         for(int i=0;i<MAX_ADDR_BITS-1;i++)
           TAPBSRtemp[IState.TAPpin2BSR(iADDR0+i)] = ADDR[i];

         // CS1 mapped for A20 for X18 DDP
         if(dutconfig.ddpranked==1)
           TAPBSRtemp[IState.TAPpin2BSR(iCSF1)] = CS_Model_[1];

         for(int i=0;i<MAX_BA_BITS;i++)
           TAPBSRtemp[IState.TAPpin2BSR(iBA0+i)] = BA[i];
         for(int i=0;i<MAX_DQ_BITS;i++)
           if ( DQ_Model[i] !== 1'bz)
             TAPBSRtemp[IState.TAPpin2BSR(iDQ0+i)] =  DQ_Model[i];
           else
             TAPBSRtemp[IState.TAPpin2BSR(iDQ0+i)] =  1'b0;  // floating DQ saved as 0
         for(int i=0;i<MAX_DQ_BITS;i++)
           if(pindrv[iDQin0+i] !== 1'bz) // model driving
             TAPBSRtemp[IState.TAPpin2BSR(iDQen0+i)] = 1'b1;
           else // tb driving
             TAPBSRtemp[IState.TAPpin2BSR(iDQen0+i)] = 1'b0;
         for(int i=0;i<MAX_DK_BITS;i++)
           TAPBSRtemp[IState.TAPpin2BSR(iDK0+i)] = DK[i];
         for(int i=0;i<MAX_DK_BITS;i++)
           TAPBSRtemp[IState.TAPpin2BSR(iDK0F+i)] = DK_[i];
         for(int i=0;i<MAX_DM_BITS;i++)
           TAPBSRtemp[IState.TAPpin2BSR(iDM0+i)] = DM[i];
         for(int i=0;i<MAX_QVLD_BITS;i++) begin
             TAPBSRtemp[IState.TAPpin2BSR(iQVLDen0+i)] = 1'b1;
             TAPBSRtemp[IState.TAPpin2BSR(iQVLD0+i)] = QVLD_Model[i];
         end
         for(int i=0;i<MAX_QK_BITS;i++)
           if (QK_Model[i] !== 1'bz)
             TAPBSRtemp[IState.TAPpin2BSR(iQK0+i)] = QK_Model[i];
           else
             TAPBSRtemp[IState.TAPpin2BSR(iQK0+i)] = 1'b0;
         for(int i=0;i<MAX_QK_BITS;i++)
           if (QK_Model[i] !== 1'bz)
             TAPBSRtemp[IState.TAPpin2BSR(iQK0F+i)] = QK_Model_[i];
           else
             TAPBSRtemp[IState.TAPpin2BSR(iQK0F+i)] = 1'b0;
         TAPBSRtemp[IState.TAPpin2BSR(iZQ)] = ZQ;

         // update to IState BSR reg
         IState.updateTAPBSRcap(TAPBSRtemp);
         
         // EXT
         // sample pins for TAP
         TAPBSREXTtemp[IState.TAPpin2BSREXT(iCK)] = CK[1];
         TAPBSREXTtemp[IState.TAPpin2BSREXT(iCKF)] = CK[0];
         TAPBSREXTtemp[IState.TAPpin2BSREXT(iCSF0)] = CS_Model_[0];
         TAPBSREXTtemp[IState.TAPpin2BSREXT(iWEF)] = WE_;
         TAPBSREXTtemp[IState.TAPpin2BSREXT(iREFF)] = REF_;
         for(int i=0;i<MAX_ADDR_BITS-1;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iADDR0+i)] = ADDR[i];

         // CS1 mapped for A20 for X18 DDP
         if(dutconfig.ddpranked==1)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iCSF1)] = CS_Model_[1];

         for(int i=0;i<MAX_BA_BITS;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iBA0+i)] = BA[i];
         for(int i=0;i<MAX_DQ_BITS;i++)
           if ( DQ_Model[i] !== 1'bz) 
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQ0+i)] =  DQ_Model[i];
           else
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQ0+i)] =  1'b0;  // floating DQ saved as 0
         for(int i=0;i<MAX_DQ_BITS;i++)
           if(pindrv[iDQin0+i] !== 1'bz) // model driving
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQen0+i)] = 1'b1;
           else // tb driving
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQen0+i)] = 1'b0;
         for(int i=0;i<MAX_DK_BITS;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iDK0+i)] = DK[i];
         for(int i=0;i<MAX_DK_BITS;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iDK0F+i)] = DK_[i];
         for(int i=0;i<MAX_DM_BITS;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iDM0+i)] = DM[i];
         for(int i=0;i<MAX_QVLD_BITS;i++) begin
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQVLDen0+i)] = 1'b1;
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQVLD0+i)] = QVLD_Model[i];
         end
         for(int i=0;i<MAX_QK_BITS;i++)
           if (QK_Model[i] !== 1'bz)
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0+i)] = QK_Model[i];
           else
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0+i)] = 1'b0;
         for(int i=0;i<MAX_QK_BITS;i++)
           if (QK_Model[i] !== 1'bz)
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0F+i)] = QK_Model_[i];
           else
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0F+i)] = 1'b0;

         for(int i=0;i<MAX_DQ_BITS;i++)
           if ( DQ_Model[i] !== 1'bz) 
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQ0+i+500)] =  DQ_Model[i];
           else
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQ0+i+500)] =  1'b0;  // floating DQ saved as 0
         for(int i=0;i<MAX_DK_BITS;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iDK0+i+500)] = DK[i];
         for(int i=0;i<MAX_DK_BITS;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iDK0F+i+500)] = DK_[i];
         for(int i=0;i<MAX_DM_BITS;i++)
           TAPBSREXTtemp[IState.TAPpin2BSREXT(iDM0+i+500)] = DM[i];
         for(int i=0;i<MAX_QVLD_BITS;i++) begin
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQVLD0+i+500)] = QVLD_Model[i];
         end
         for(int i=0;i<MAX_QK_BITS;i++)
           if (QK_Model[i] !== 1'bz)
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0+i+500)] = QK_Model[i];
           else
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0+i+500)] = 1'b0;
         for(int i=0;i<MAX_QK_BITS;i++)
           if (QK_Model[i] !== 1'bz)
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0F+i+500)] = QK_Model_[i];
           else
             TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0F+i+500)] = 1'b0;

         TAPBSREXTtemp[IState.TAPpin2BSREXT(iZQ)] = ZQ;

         // update to IState BSR reg
         IState.updateTAPBSREXTcap(TAPBSREXTtemp);
         
         // update TAP state
         IState.updateTAPtable(TMS, TDI, tdo, tdo_valid);
         
         // update from BSR to pins
         if (dutstate.curTAPIRstate == HIGH_Z)  begin
            tapon=1'b1;
            tapdrv[iQVLD1:iQVLD0]='z;
            tapdrv[iQK3:iQK0]='z;
            tapdrv[iQK3F:iQK0F]='z;
            tapdrv[iDQin35:iDQin0]={36{1'bz}};
         end
         else if (dutstate.curTAPIRstate == CLAMP || dutstate.curTAPIRstate == EXTEST) begin
            tapon=1'b1;
            TAPBSRtemp = IState.getTAPBSR();
            for(int i=0;i<MAX_QVLD_BITS;i++)
              tapdrv[iQVLD0+i]=TAPBSRtemp[IState.TAPpin2BSR(iQVLD0+i)];
            // ddp x18 QVLD1 is not connected
            if(dutconfig.archtype == RLDRAM3DDP && dutconfig.ddpranked==1)
              tapdrv[iQVLD0+1] = 1'bz;
            for(int i=0;i<MAX_QK_BITS;i++)
              tapdrv[iQK0+i]=TAPBSRtemp[IState.TAPpin2BSR(iQK0+i)];
            for(int i=0;i<MAX_QK_BITS;i++)
              tapdrv[iQK0F+i]=TAPBSRtemp[IState.TAPpin2BSR(iQK0F+i)];
            for(int i=0;i<MAX_DQ_BITS;i++)
              if(TAPBSRtemp[IState.TAPpin2BSR(iDQen0+i)]===1'b1)
                // DDP X18 upper dq always hiz
                if (i>=18 && dutconfig.archtype==RLDRAM3DDP && dutconfig.ddpranked==1)
                  tapdrv[iDQin0+i]=1'bz;
                else
                  tapdrv[iDQin0+i]=TAPBSRtemp[IState.TAPpin2BSR(iDQ0+i)];
              else
                tapdrv[iDQin0+i]=1'bz;
         end
         else if (dutstate.curTAPIRstate == EXTEST_EXT) begin
            tapon=1'b1;
            TAPBSREXTtemp = IState.getTAPBSREXT();
            for(int i=0;i<MAX_QVLD_BITS;i++)
              tapdrv[iQVLD0+i]=TAPBSREXTtemp[IState.TAPpin2BSREXT(iQVLD0+i)];
            // ddp x18 QVLD1 is not connected
            if(dutconfig.archtype == RLDRAM3DDP && dutconfig.ddpranked==1)
              tapdrv[iQVLD0+1] = 1'bz;
            for(int i=0;i<MAX_QK_BITS;i++)
              tapdrv[iQK0+i]=TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0+i)];
            for(int i=0;i<MAX_QK_BITS;i++)
              tapdrv[iQK0F+i]=TAPBSREXTtemp[IState.TAPpin2BSREXT(iQK0F+i)];
            for(int i=0;i<MAX_DQ_BITS;i++)
              if(TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQen0+i)]===1'b1)
                // DDP X18 upper dq always hiz
                if (i>=18 && dutconfig.archtype==RLDRAM3DDP && dutconfig.ddpranked==1)
                  tapdrv[iDQin0+i]=1'bz;
                else
                  tapdrv[iDQin0+i]=TAPBSREXTtemp[IState.TAPpin2BSREXT(iDQ0+i)];
              else
                tapdrv[iDQin0+i]=1'bz;
         end
         else
           tapon=1'b0;
         
      end
      else begin // falling edge
         if (tdo_valid)
           pindrv[iTDO] <= #2500ps tdo;
         else
           pindrv[iTDO] <= #2500ps 1'bz;
      end
   end

   // Load needed timing
   function void updatetiming;
      input                      UTYPE_sharedtiming ts;
      begin
         timing=ts;
         timing.tOffset = tCK;
      end
   endfunction 
   
endmodule

