/****************************************************************************************
*
*   Disclaimer   This software code and all associated documentation, comments or other
*  of Warranty:  information (collectively "Software") is provided "AS IS" without
*                warranty of any kind. MICRON TECHNOLOGY, INC. ("MTI") EXPRESSLY
*                DISCLAIMS ALL WARRANTIES EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
*                TO, NONINFRINGEMENT OF THIRD PARTY RIGHTS, AND ANY IMPLIED WARRANTIES
*                OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE. MTI DOES NOT
*                WARRANT THAT THE SOFTWARE WILL MEET YOUR REQUIREMENTS, OR THAT THE
*                OPERATION OF THE SOFTWARE WILL BE UNINTERRUPTED OR ERROR-FREE.
*                FURTHERMORE, MTI DOES NOT MAKE ANY REPRESENTATIONS REGARDING THE USE OR
*                THE RESULTS OF THE USE OF THE SOFTWARE IN TERMS OF ITS CORRECTNESS,
*                ACCURACY, RELIABILITY, OR OTHERWISE. THE ENTIRE RISK ARISING OUT OF USE
*                OR PERFORMANCE OF THE SOFTWARE REMAINS WITH YOU. IN NO EVENT SHALL MTI,
*                ITS AFFILIATED COMPANIES OR THEIR SUPPLIERS BE LIABLE FOR ANY DIRECT,
*                INDIRECT, CONSEQUENTIAL, INCIDENTAL, OR SPECIAL DAMAGES (INCLUDING,
*                WITHOUT LIMITATION, DAMAGES FOR LOSS OF PROFITS, BUSINESS INTERRUPTION,
*                OR LOSS OF INFORMATION) ARISING OUT OF YOUR USE OF OR INABILITY TO USE
*                THE SOFTWARE, EVEN IF MTI HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
*                DAMAGES. Because some jurisdictions prohibit the exclusion or
*                limitation of liability for consequential or incidental damages, the
*                above limitation may not apply to you.
*
*                Copyright 2003 Micron Technology, Inc. All rights reserved.
*
****************************************************************************************/

`timescale 1ps/1ps

`ifndef RLDRAM3_PACKAGE
 `define RLDRAM3_PACKAGE

package rldram3_package;
   
   // ==================================================================================
   // max sizes of the parts

   // These 2 are moved outside package to support parameterized module
   // needs to be added to all modules that imports rldram3_package
   // parameter MAX_DQ_BITS = 36;
   // parameter MAX_QK_BITS = 4;
   
   // normal functions
`ifdef rldram3_1Gb
   parameter MAX_ADDR_BITS = 22;  // add 1 bit for DDP
`else // 576Mb
   parameter MAX_ADDR_BITS = 21;  // add 1 bit for DDP
`endif
   parameter MAX_BA_BITS = 4;
   parameter MAX_MR_BITS = 24; // 4 MSB is MR bank
   parameter MAX_DK_BITS = 2;
   parameter MAX_BURST_LEN = 8;
   parameter MAX_DM_BITS = 2;
   parameter MAX_QVLD_BITS = 2;
   parameter MAX_BANKS = 2**MAX_BA_BITS;

   // max REFM banks allowed
   parameter MAX_REFM_BANKS = 4;

   // ==================================================================================
   // arch params
   typedef   enum { RLDRAM3, RLDRAM3DDP } UTYPE_archtype;	 

   // BL type
   typedef   enum {BSTL2=2, BSTL4=4, BSTL8=8 } UTYPE_carchtype;
   
   // ==================================================================================
   // mode registers
   parameter MAX_MR_REG = 3;
   
   // Mode Register Definitions
   //                   BA   A19  A15  A11  A7   A3
   parameter MR0  = 24'b0000_0000_0000_0000_0000_0000;
   parameter MR1  = 24'b0001_0000_0000_0000_0000_0000;
   parameter MR2  = 24'b0010_0000_0000_0000_0000_0000;

   function void checkMRBA(int ba);
      if (ba<0 || ba>2)
	$display("ERROR %m @ %t MR bank address out of range", $time);
   endfunction
   
   // MR0 Codes

   // tRC
   parameter MR0_tRC_MASK = 24'b0000_0000_0000_0000_0000_1111;

   function [MAX_MR_BITS-1:0] tRC2MR(int tRC);
      checktRC(tRC);
      tRC2MR = tRC-2;
   endfunction

   function int MR2tRC(reg [MAX_MR_BITS-1:0] mr);
      MR2tRC = (mr&MR0_tRC_MASK)+2;
      checktRC(MR2tRC);
   endfunction
   
   function void checktRC(int tRC);
      if (tRC<2 || tRC>15) 
	$display("ERROR %m @ %t tRC %0d out of range", $time, tRC);
   endfunction
   
   // DL
   parameter MR0_DL_MASK = 24'b0000_0000_0000_0000_1111_0000;
   
   function [MAX_MR_BITS-1:0] DL2MR(int dl);
      checkDL(dl);
      DL2MR = MR0_DL_MASK & ((dl-3)<<4);
   endfunction
   
   function int MR2DL(reg [MAX_MR_BITS-1:0] mr);
      MR2DL = ((mr&MR0_DL_MASK)>>4)+3;
      checkDL(MR2DL);
   endfunction
   
   function void checkDL(int dl);
      if (dl<3 || dl>18 )
	$display("ERROR %m @ %t DL %0d out of range", $time, dl);
   endfunction

   // DLL
   function [MAX_MR_BITS-1:0] DLL2MR(reg in);
      DLL2MR = in<<8;
   endfunction
   
   function reg MR2DLL(reg [MAX_MR_BITS-1:0] mr);
      MR2DLL = (mr>>8) & 24'b1;
   endfunction

   // AM
   function [MAX_MR_BITS-1:0] AM2MR(reg in);
      AM2MR = in<<9;
   endfunction
   
   function reg MR2AM(reg [MAX_MR_BITS-1:0] mr);
      MR2AM = (mr>>9) & 24'b1;
   endfunction
   
   // MR1 Codes

   // BL
   parameter MR1_BL2 = 24'b0000_0000_0000_0000_0000_0000;
   parameter MR1_BL4 = 24'b0000_0000_0000_0010_0000_0000;
   parameter MR1_BL8 = 24'b0000_0000_0000_0100_0000_0000;
   parameter MR1_BL_MASK = 24'b0000_0000_0000_0110_0000_0000;
   
   function [MAX_MR_BITS-1:0] BL2MR(int bl);
      checkBL(bl);
      case (bl)
	2: BL2MR = MR1_BL2;
	4: BL2MR = MR1_BL4;
	8: BL2MR = MR1_BL8;
      endcase
   endfunction

   function int MR2BL(reg [MAX_MR_BITS-1:0] mr);
      MR2BL = 2**(((mr&MR1_BL_MASK)>>9)+1);
      checkBL(MR2BL);
   endfunction
   
   function void checkBL(int bl);
      if (bl!==2 && bl!==4 && bl!==8) 
	$display("ERROR %m @ %t BL %0d out of range", $time, bl);
   endfunction
   
   // ODI
   parameter MR1_ODI40  = 24'b0000_0000_0000_0000_0000_0000;
   parameter MR1_ODI60  = 24'b0000_0000_0000_0000_0000_0001;
   parameter MR1_ODIRES = 24'b0000_0000_0000_0000_0000_0010;
   parameter MR1_ODI240 = 24'b0000_0000_0000_0000_0000_0011;
   parameter MR1_ODI_MASK = 24'b0000_0000_0000_0000_0000_0011;
   
   function [MAX_MR_BITS-1:0] ODI2MR(reg [2:0] odi);
      case (odi)
	0:  ODI2MR = MR1_ODI40;
	1:  ODI2MR = MR1_ODI60;
	2:  ODI2MR = MR1_ODIRES;
	3:  ODI2MR = MR1_ODI240;
	default: ODI2MR = MR1_ODI40;
      endcase
   endfunction
   
   function reg [2:0] MR2ODI(reg [MAX_MR_BITS-1:0] mr);
      case (mr&MR1_ODI_MASK)
	MR1_ODI40:  MR2ODI=0;
	MR1_ODI60:  MR2ODI=1;
	MR1_ODIRES: MR2ODI=2;
	MR1_ODI240: MR2ODI=3;
	default: MR2ODI=0;
      endcase
   endfunction
   
   // ODT
   parameter MR1_ODT0   = 24'b0000_0000_0000_0000_0000_0000;
   parameter MR1_ODT40  = 24'b0000_0000_0000_0000_0000_0100;
   parameter MR1_ODT60  = 24'b0000_0000_0000_0000_0000_1000;
   parameter MR1_ODT120 = 24'b0000_0000_0000_0000_0000_1100;
   parameter MR1_ODT_MASK = 24'b0000_0000_0000_0000_0001_1100;
   
   function [MAX_MR_BITS-1:0] ODT2MR(reg [2:0] odt);
      case (odt)
	0:  ODT2MR = MR1_ODT0;
	1:  ODT2MR = MR1_ODT40;
	2:  ODT2MR = MR1_ODT60;
	3:  ODT2MR = MR1_ODT120;
	default: ODT2MR = MR1_ODT0;
      endcase
   endfunction
   
   function reg [2:0] MR2ODT(reg [MAX_MR_BITS-1:0] mr);
      case (mr&MR1_ODT_MASK)
	MR1_ODT0:   MR2ODT=0;
	MR1_ODT40:  MR2ODT=1;
	MR1_ODT60:  MR2ODT=2;
	MR1_ODT120: MR2ODT=3;
	default: MR2ODT=0;
      endcase
   endfunction
   
   // DLLRESET
   function [MAX_MR_BITS-1:0] DLLRESET2MR(reg in);
      DLLRESET2MR = (in<<5) ;
   endfunction
   
   function reg MR2DLLRESET(reg [MAX_MR_BITS-1:0] mr);
      MR2DLLRESET = (mr>>5) & 24'b1;
   endfunction

   // ZQ
   function [MAX_MR_BITS-1:0] ZQ2MR(reg in);
      ZQ2MR = (in<<6) ;
   endfunction
   
   function reg MR2ZQ(reg [MAX_MR_BITS-1:0] mr);
      MR2ZQ = (mr>>6) & 24'b1;
   endfunction
   
   // ZQe
   function [MAX_MR_BITS-1:0] ZQe2MR(reg in);
      ZQe2MR = (in<<7) ;
   endfunction
   
   function reg MR2ZQe(reg [MAX_MR_BITS-1:0] mr);
      MR2ZQe = (mr>>7) & 24'b1;
   endfunction
   
   // REF
   function [MAX_MR_BITS-1:0] REF2MR(reg in);
      REF2MR = (in<<8) ;
   endfunction
   
   function reg MR2REF(reg [MAX_MR_BITS-1:0] mr);
      MR2REF = (mr>>8) & 24'b1;
   endfunction

   //ODT Relinquish
   function [MAX_MR_BITS-1:0] ODTREL2MR(reg [1:0] in);
      ODTREL2MR = (in<<12) ;
   endfunction
   
   function reg [1:0] MR2ODTREL(reg [MAX_MR_BITS-1:0] mr);
      MR2ODTREL = (mr>>12) & 24'b11;
   endfunction
   
   // MR2 Codes // mainly for TRAINING
   // RT
   function [MAX_MR_BITS-1:0] RT2MR(reg in);
      RT2MR = (in<<2) | MR2;
   endfunction
   
   function reg MR2RT(reg [MAX_MR_BITS-1:0] mr);
      MR2RT = (mr>>2) & 24'b1;
   endfunction

   // RTREG
   function [MAX_MR_BITS-1:0] RTREG2MR(reg [1:0] in);
      RTREG2MR = (in) | MR2;
   endfunction
   
   function reg [1:0] MR2RTREG(reg [MAX_MR_BITS-1:0] mr);
      MR2RTREG = (mr) & 24'b11;
   endfunction

   // WRITE
`ifndef rldram3_1Gb
   parameter MR2_WRITE_MASK = 24'b0000_0000_0000_0000_0001_1000;
`else
   parameter MR2_WRITE_MASK = 24'b0000_0000_0000_0000_0011_1000;
`endif   
   
   function [MAX_MR_BITS-1:0] WRITE2MR(int wr);
      checkWRITE(wr);
      WRITE2MR = MR2_WRITE_MASK & (wr<<3) ;
   endfunction
   
   function int MR2WRITE(reg [MAX_MR_BITS-1:0] mr);
      MR2WRITE = (mr&MR2_WRITE_MASK)>>3;
      checkWRITE(MR2WRITE);
   endfunction
   
   function void checkWRITE(int in);
`ifndef rldram3_1Gb
      if (in<0 || in>2 )
`else
      if (in<0 || in>4 )
`endif
	$display("ERROR %m @ %t WRITE out of range", $time);
   endfunction

   function int numbankWRITE(int in);
      case(in)
        0: numbankWRITE = 1;
        1: numbankWRITE = 2;
        2: numbankWRITE = 4;
        3: numbankWRITE = 2;
        4: numbankWRITE = 4;
      endcase
   endfunction

   function int banklWRITE(int in, int ind, int bank);
      case (in)
        0: banklWRITE = bank;
        1: banklWRITE = bank^(ind<<3);
        2: banklWRITE = bank^(ind<<2);
        3: banklWRITE = (bank+ind)%16;
        4: banklWRITE = (bank+ind)%16;
      endcase
   endfunction
   
   // ==================================================================================
   // JTAG TAP
   parameter INST_LEN   = 8;
   parameter BYPASS_LEN = 1;
   parameter ID_LEN     = 32;
`ifdef rldram3_1Gb
   parameter BSR_LEN    = 135;
   parameter BSR_LEN_EXT = 187;
`else
   parameter BSR_LEN    = 121;
   parameter BSR_LEN_EXT = 173; 
`endif
   
   // TAP state
   typedef   enum {TAPreset, TAPidle, TAPDRscan, TAPIRscan, TAPCaptureDR, TAPCaptureIR,
		   TAPShiftDR, TAPShiftIR, TAPExit1DR, TAPExit1IR, TAPPauseDR, TAPPauseIR,
		   TAPExit2DR, TAPExit2IR, TAPUpdateDR, TAPUpdateIR} UTYPE_taptype;
   
   typedef   enum {
                   EXTEST	 = 32'b0000_0000,
                   EXTEST_EXT	 = 32'b0000_1000,
		   ID_CODE	 = 32'b0010_0001,
		   SAMPLE_PRELOAD= 32'b0000_0101,
		   SAMPLE_PRELOAD_EXT=32'b0000_1101,
		   CLAMP 	 = 32'b0000_0111,
		   HIGH_Z	 = 32'b0000_0011,
		   MR0_MR1	 = 32'b0010_0101,
		   MR2_QED_TS	 = 32'b0010_1001,
		   FUSEID1   	 = 32'b0010_1101,
		   FUSEID2   	 = 32'b0011_0001,
		   INTCNTRL1 	 = 32'b0011_0101,
		   INTCNTRL2 	 = 32'b0011_1001,
		   BYPASS	 = 32'b1111_1111,
		   TAPUNKOWN     = 32'b1_0000_0000} UTYPE_tapirtype;
   
   // ==================================================================================
   // shared timing block

   typedef   struct packed{
			   int tOffset;
			   
			   // spec timing
			   int tCK, tCK_min, tCK_max;
			   int ClockDutyCycle;
			   int DataClockDutyCycle;
			   int tIS, tIH, tIPW;
			   int tCKDK, tCKQK, tCKQK_DLL_DIS, tRC, tQKVLD, tQKQ;
			   int tDS, tDH, tDIPW ;
			   int tDLLkc, tMRSCc, tMMDc;
			   int EXP_Advance;
			   } UTYPE_sharedtiming;

   // ==================================================================================
   // fixed (can not change on the fly)  configuration of the chip
   typedef   enum {BYX36=36, BYX18=18} UTYPE_bymode;

   typedef   struct packed{
			   UTYPE_archtype archtype;
			   UTYPE_bymode config_bymode;
			   UTYPE_bymode bymode;
			   int numbank, numdq, numqk, numdm, numqvld, numaddr;
    reg      ddpranked; // DDP 2CS option
    reg      sio; // SIO option
    reg [MAX_ADDR_BITS-1:0] addrmask;
    reg [MAX_BA_BITS-1:0]   bankaddrmask;
    //reg [MAX_DQ_BITS-1:0]   dqmask;
    //reg [MAX_QK_BITS-1:0]   qkmask;  // used hard coded max width
    reg [36-1:0] 	    dqmask;  
    reg [4-1:0] 	    qkmask;
    reg [31:0] 		    JTAG_IDCODE;
			   } UTYPE_dutconfig;			   

   // ==================================================================================
   // cmd state 

   typedef 		   enum {ST_RESET, ST_SPEC, ST_TRAIN
				 } UTYPE_cmdstate;
   
   typedef 		   enum{
				// real commands
				cmdINIT, cmdWR, cmdWRx, cmdWRy, cmdRD, cmdRDx, cmdRDy, cmdRESET, 
				cmdNOP, cmdDES, 
				cmdLMR, cmdLMRx, cmdLMRy, cmdLMRDLLRESET,
				cmdREF1, cmdREF1x, cmdREF1y, cmdREFM, cmdREFMx, cmdREFMy,
				cmdRT, cmdRTx, cmdRTy, cmd_CCMD,
				cmdNUM
				} UTYPE_cmdtype;
   
   typedef 		   struct packed {
					  UTYPE_cmdtype cmd;
    reg [MAX_BA_BITS-1:0]  bank;
    reg [MAX_ADDR_BITS-1:0] addr, addry;
    reg 		    addryflag, cs;
					  int tCK;
 					  } UTYPE_cmdpacket;

   // ==================================================================================
   // state of the chip
   typedef 		    struct packed {
					   // MR0
    integer 		    BL, DL, tRC; 
    reg 		    DLL, AM;
					   
					   // MR1
    reg [1:0] 		    ODI, ODTREL;
    reg [2:0]  		    ODT;
    reg 		    DLLRESET, REF, ZQe, ZQ;
					   
					   // MR2
    reg [2:0] 		    WRITE;
    reg 		    RT;
    reg [1:0] 		    RTreg;

					   // TMflag, no corresponding register
    reg 		    TMLAT3;
					   } UTYPE_dutmodeconfig;


   typedef 		    struct packed {
					   int 		   cyclecount;
					   UTYPE_cmdstate cmdstate;
					   UTYPE_cmdtype cmd;
					   int cmdextranops;
					   int cmdbank;
					   int cmdaddr;
					   int cmdtCK;
					   int bladdrmask;
					   
					   // TAP
					   UTYPE_taptype curTAPstate;
					   UTYPE_tapirtype curTAPIRstate;

    integer		    dly1, dly2, dly3;
    bit 		    qkvalid;
    time 		    cycletime;				  
    bit [2*MAX_BANKS-1:0]   activebank, refbank;
    integer 		    Write_Latency, Read_Latency; 		    
					   
					   } UTYPE_dutstate; 			   

   // address
   typedef reg [MAX_BA_BITS-1:0] UTYPE_bankaddr;
   typedef reg [MAX_ADDR_BITS-1:0] UTYPE_addr;

   // Random 2 bank addr structure
   typedef 			   struct packed {
                                                  UTYPE_bankaddr Bank_A0i;
                                                  UTYPE_bankaddr Bank_A1i;
                                                  UTYPE_addr  A0i;
                                                  UTYPE_addr  A1i;
                                                  } UTYPE_randaddr;

   // ==================================================================================
   // type of each pin
   parameter MAX_PINV = 4;
   
   typedef   enum { CKPIN, ADDRPIN, DQPIN } UTYPE_PINTYPE;
   
   // array of pin index
   // order is important, do not change
   typedef   enum {iCK, iCKF, iCSF0, iCSF1, iWEF, iREFF,
		   iADDR0, iADDR1, iADDR2, iADDR3, iADDR4, iADDR5, iADDR6, iADDR7, iADDR8, iADDR9, iADDR10, 
		   iADDR11, iADDR12, iADDR13, iADDR14, iADDR15, iADDR16, iADDR17, iADDR18, iADDR19, iADDR20, 
`ifdef rldram3_1Gb
		   iADDR21, 
`endif
		   iBA0, iBA1, iBA2, iBA3,
		   iDQ0, iDQ1, iDQ2, iDQ3, iDQ4, iDQ5, iDQ6, iDQ7, iDQ8,
		   iDQ9, iDQ10, iDQ11, iDQ12, iDQ13, iDQ14, iDQ15, iDQ16, iDQ17,
		   iDQ18, iDQ19, iDQ20, iDQ21, iDQ22, iDQ23, iDQ24, iDQ25, iDQ26,
		   iDQ27, iDQ28, iDQ29, iDQ30, iDQ31, iDQ32, iDQ33, iDQ34, iDQ35,
		   iDQin0, iDQin1, iDQin2, iDQin3, iDQin4, iDQin5, iDQin6, iDQin7, iDQin8,
		   iDQin9, iDQin10, iDQin11, iDQin12, iDQin13, iDQin14, iDQin15, iDQin16, iDQin17,
		   iDQin18, iDQin19, iDQin20, iDQin21, iDQin22, iDQin23, iDQin24, iDQin25, iDQin26,
		   iDQin27, iDQin28, iDQin29, iDQin30, iDQin31, iDQin32, iDQin33, iDQin34, iDQin35,
		   iDQen0, iDQen1, iDQen2, iDQen3, iDQen4, iDQen5, iDQen6, iDQen7, iDQen8,
		   iDQen9, iDQen10, iDQen11, iDQen12, iDQen13, iDQen14, iDQen15, iDQen16, iDQen17,
		   iDQen18, iDQen19, iDQen20, iDQen21, iDQen22, iDQen23, iDQen24, iDQen25, iDQen26,
		   iDQen27, iDQen28, iDQen29, iDQen30, iDQen31, iDQen32, iDQen33, iDQen34, iDQen35,
		   iDM0, iDM1, iDK0, iDK1, iDK0F, iDK1F,
		   iQVLD0, iQVLD1, iQK0, iQK1, iQK2, iQK3, iQK0F, iQK1F, iQK2F, iQK3F,
		   iQVLDen0, iQVLDen1, iQKen0, iQKen1, iQKen2, iQKen3, iQKen0F, iQKen1F, iQKen2F, iQKen3F,
		   iZQ,iRESET,iMF,
		   iTDO, iTMS, iTDI, iTCK,
		   iPWR, iVREF, iPRBPAD, iDVC2, iVSS0, iVSS1, iVSS2, iVSS3,
		   iMAXPINS} UTYPE_pinindex;

   // clkL_H commands
   parameter CNOP    =  3'b100;
   parameter CLMR    =  3'b000;
   parameter CREAD   =  3'b011;
   parameter CWRITE  =  3'b001;
   parameter CREF    =  3'b010;
   
endpackage

`endif
